import numpy as np
from gym import spaces

class PGPE(object):
    """
    Agent implementing PGPE with arbitrary policy (symmetric version)
    eval_func: function from policy_params to reward
    """
    def __init__(self, eval_func, policy_params, **userconfig):

        self.eval_func  = eval_func

        self.policy_params = policy_params

        self.config = {
            "init_mean" : 0.0,      # Initialize meta-policy parameters with this mean
            "init_std" : 1.0,       # Initialize meta-policy parameters with this standard deviation
            "learning_rate" : 0.01,
            "learning_rate_sigma" : 0.01,
            "record_every" : 10,
            "max_reward" : 0,        # known maximal reward
            "baseline_avg" : 0.05,
            "name" : "pgpe"
        }
        self.config.update(userconfig)
        self.theta_n = len(self.policy_params)
        print ("num parameters", self.theta_n)
        self.mu    = np.zeros(self.theta_n)  + self.config["init_mean"]
        self.sigma = np.ones(self.theta_n) * self.config["init_std"]

        # parameter buffer
        self.param_buffer = []

        # episode Buffer
        self.mean_reward=0.0
        self.std_reward=1.0
        self.max_reward = self.config["max_reward"]
        self.baseline_avg = self.config["baseline_avg"]
        # parameters
        self.lr = self.config["learning_rate"]
        self.lr_sigma = self.config["learning_rate_sigma"]
        self.t=0
        self.episode_cnt=0

    def train(self, epochs_n, monitor=None):

        for e in range(epochs_n/2):
            eps = np.random.randn(self.theta_n)*self.sigma
            theta1 = self.mu + eps
            theta2 = self.mu - eps
            (r1,seed)     = self.eval_func(e, theta1, seed=None)
            (r2,_)        = self.eval_func(e, theta2, seed=seed)
            rate = 1.0/(e+2) + self.baseline_avg
            self.mean_reward = self.mean_reward * (1.0 - rate) + (r1 + r2) / 2.0 * rate
            self.max_reward = np.max([r1 + 0.1, r2 + 0.1, self.max_reward])
            self.mu    += eps * (self.lr * (r1 - r2) / (2.0 * self.max_reward - r1 - r2))
            update = ((np.square(eps) - np.square(self.sigma)) / self.sigma) * \
                          (self.lr_sigma / (self.max_reward - self.mean_reward) * \
                           ((r1 + r2) / 2.0-self.mean_reward))
            self.sigma += update
            self.sigma = np.clip(self.sigma, 0.05, 10.0)
            if monitor is not None:
                    monitor(self)
            if e % self.config["record_every"] == 0:
                self.param_buffer.append((self.mu.copy(), self.sigma.copy()))
            self.episode_cnt +=2

    def test_run(self, eval_fun = None, with_sampling=False):
        f = self.eval_fun if eval_fun is None else eval_fun
        if with_sampling:
            theta = self.mu + np.random.randn(self.theta_n)*self.sigma
        else:
            theta = self.mu
        return f(1,theta)
