import numpy as np
import scipy.sparse

################################################################################
#  R B F
################################################################################

# radial basis function with mean "mu" and width "sigma", which can be used with tensors for mu (and also x)
def rbf(x,mu,sigma):
    d = x-mu
    return np.exp(-np.sum(d*d/(2*sigma*sigma), axis=1))
# tent basis function with mean "mu" and width "sigma", which can be used with tensors for mu (and also x)
def tent(x,mu,sigma):
    d = x-mu
    return np.max(1.0-np.abs(d)/sigma,0.0)


# this function gets the observations space and returns a function that will calculate the
#  feature activations for a given observation
def rbf_featuremap(o_space, features_per_dim, limits=None):
    dim = o_space.shape[0]
    if dim != len(features_per_dim):
        raise ValueError("features_per_dim should be list of len " + str(dim))
    if limits is None:
        limits = zip(o_space.low, o_space.high)
    limits = np.asarray(limits)
    min_coord = limits[:,0]
    width     = limits[:,1] - min_coord
    gridpoints = [np.linspace(0, 1, d) for d in features_per_dim]
    grid = np.transpose(np.meshgrid(*gridpoints)).reshape((-1,dim))
    size = 1./np.asarray(features_per_dim)
    def feature_fun(ob):
        arr= rbf((np.asarray(ob) - min_coord)/width, grid, size)
        arr[-1]=.1; # set last feature to constant
        return arr
    return feature_fun

################################################################################
#  R B F  component wise
################################################################################

# radial basis function with mean "mu" and width "sigma", which can be used with tensors for mu (and also x)
def rbf_comp(x,mu,sigma):
    d = x-mu
    return np.exp(-d*d/(2*sigma*sigma))

# this function gets the observations space and returns a function that will calculate the
#  feature activations for a given observation
def rbf_comp_featuremap(o_space, features_per_dim=5, limits=None):
    dim = o_space.shape[0]
    if limits is None:
        limits = zip(o_space.low, o_space.high)
    limits = np.asarray(limits)
    min_coord = limits[:,0]
    width     = limits[:,1] - min_coord
    centers = (np.linspace(0, 1, features_per_dim)).reshape(-1,1)
    size = 1./features_per_dim
    def feature_fun(ob):
        arr= rbf_comp((np.asarray(ob) - min_coord)/width, centers, size).reshape(-1)
        arr[-1]=1; # set last feature to constant
        return arr
    return feature_fun


################################################################################
# random sparse features
################################################################################

# this function gets the observations space and returns a function that will calculate the
#  feature activations for a given observation
def random_featuremap(o_space, feature_number=600, limits=None, weights=None, strength=1.0, sparsity=0.4):
    dim = o_space.shape[0]
    if limits is None:
        limits = zip(o_space.low, o_space.high)
    limits = np.asarray(limits)
    min_coord = limits[:,0]
    width     = limits[:,1] - min_coord

    if weights is None:
        weights   = np.asarray(scipy.sparse.rand(dim+1,feature_number,sparsity).todense());
        weights*= strength*np.random.randint(low=-1,high=2, size=weights.shape)

    def feature_fun(ob):
        arr= np.clip(np.dot(np.hstack(((np.asarray(ob) - min_coord)/width,[1.0])), weights),0.0,1.0)
        arr[-1]=1; # set last feature to constant
        return arr
    return weights,feature_fun
