from gym import spaces
import numpy as np

class LinearQAgent(object):
    """
    Agent implementing Q-learning with linear function approximation.
    """
    def __init__(self, observation_space, action_space, featuremap, **userconfig):
        if not isinstance(observation_space, spaces.box.Box):
            raise UnsupportedSpace('Observation space {} incompatible with {}. (Require: Box)'.format(observation_space, self))
        if not isinstance(action_space, spaces.discrete.Discrete):
            raise UnsupportedSpace('Action space {} incompatible with {}. (Reqire Discrete.)'.format(action_space, self))
        self.observation_space = observation_space
        self.action_space = action_space
        self.action_n = action_space.n
        self.featuremap = featuremap
        self.config = {
            "init_mean" : 0.0,      # Initialize Q values with this mean
            "init_std" : 0.0,       # Initialize Q values with this standard deviation
            "learning_rate" : 0.01,
            "eps": 0.05,            # Epsilon in epsilon greedy policies
            "discount": 0.95,
            "buffersize" : 10000,
            "replay" : 0,
            "only_store_good_runs" : False
        }
        self.config.update(userconfig)
        inputdimension=observation_space.low.shape
        ob = np.random.randn(*inputdimension)
        features = featuremap(ob)
        self.feature_n = len(features)
        print ("num features", self.feature_n)
        self.W = np.random.random((self.action_n, self.feature_n)) * self.config["init_std"]  + self.config["init_mean"]
        # experience Buffer
        self.buffer = []
        self.replay=self.config["replay"]
        self.buffersize=self.config["buffersize"]
        self.episode_buffer = []
        self.episode_reward=0.0
        self.mean_reward=0.0
        self.std_reward=1.0
        # parameters
        self.lr = self.config["learning_rate"]
        self.discount = self.config["discount"]
        self.eps = self.config["eps"]
        self.t=0
        self.episode_cnt=0

    def reset(self):
        n = self.episode_cnt
        if self.replay>0 and self.t>0:
            if not self.config["only_store_good_runs"] or (self.episode_reward > (self.mean_reward+np.sqrt(self.std_reward))):
                #print "adding to buffer: ", self.episode_buffer, n
                for entry in self.episode_buffer:
                    if len(self.buffer)<self.buffersize:
                        self.buffer.append(entry)
                    else:
                        self.buffer[self.t%self.buffersize] = entry
                #print "add episode: ", n, "reward", self.episode_reward, \
                #        "mean", self.mean_reward, "std", np.sqrt(self.std_reward), " len", len(self.buffer)

        self.mean_reward = (self.episode_reward + n *self.mean_reward)/(n+1)
        if n>0:
            self.std_reward = (n-1.0)/(n)*self.std_reward + ((self.episode_reward - self.mean_reward)**2)/(n+1)

        self.episode_cnt+=1
        self.episode_reward=0.0
        self.episode_buffer=[]

    def getQ(self,observation,action):
        features = self.featuremap(observation)
        return np.dot(self.W[action],features)

    def getQ_(self,features,action):
        return np.dot(self.W[action],features)

    def act(self, observation, eps=None):
        if eps is None:
            eps = self.eps

        # epsilon greedy.
        if np.random.random() > eps:
            action = np.argmax([self.getQ(observation, a) for a in range(self.action_n)])
        else:
            action = self.action_space.sample()
        return action

    def update(self, ob, action, reward, ob_next):
        features = self.featuremap(ob)
        features_next = self.featuremap(ob_next)
        V_future = np.max([self.getQ_(features_next, a) for a in range(self.action_n)])
        self.W[action] += self.lr * \
                (reward + self.discount * V_future - self.getQ_(features, action))*features

    def learn(self, ob, action, reward, ob_next):
        self.update(ob,action,reward,ob_next)
        self.t+=1
        self.episode_reward += reward
        if self.replay>0:
            self.episode_buffer.append((ob,action,reward,ob_next))
            blen=len(self.buffer)
            # sample and learn
            if blen>0:
                samples=np.random.randint(blen,size=min(self.replay,blen))
                for s in samples:
                    self.update(*self.buffer[s])
